@@ -6,8 +6,8 @@
 /*                                                                           */
 /*                                                                           */
 /*                                                                           */
-/* (C) 1998-2003 Ullrich von Bassewitz                                       */
-/*               Rmerstrae 52                                              */
+/* (C) 1998-2008 Ullrich von Bassewitz                                       */
+/*               Roemerstrasse 52                                            */
 /*               D-70794 Filderstadt                                         */
 /* EMail:        uz@cc65.org                                                 */
 /*                                                                           */
@@ -46,6 +46,7 @@
 #include "condasm.h"
 #include "error.h"
 #include "global.h"
+#include "instr.h"
 #include "istack.h"
 #include "nexttok.h"
 #include "pseudo.h"
@@ -87,7 +88,7 @@ static int HT_Compare (const void* Key1, const void* Key2);
 typedef struct IdDesc IdDesc;
 struct IdDesc {
     IdDesc*  	    Next;     	/* Linked list */
-    char       	    Id [1];	/* Identifier, dynamically allocated */
+    StrBuf          Id;         /* Identifier, dynamically allocated */
 };
 
 
@@ -105,7 +106,7 @@ struct Macro {
     TokNode* 	    TokRoot;	/* Root of token list */
     TokNode* 	    TokLast;	/* Pointer to last token in list */
     unsigned char   Style;	/* Macro style */
-    char       	    Name [1];	/* Macro name, dynamically allocated */
+    StrBuf          Name;	/* Macro name, dynamically allocated */
 };
 
 /* Hash table functions */
@@ -156,7 +157,7 @@ static unsigned LocalName = 0;
 static unsigned HT_GenHash (const void* Key)
 /* Generate the hash over a key. */
 {
-    return HashStr (Key);
+    return HashBuf (Key);
 }
 
 
@@ -164,7 +165,7 @@ static unsigned HT_GenHash (const void* Key)
 static const void* HT_GetKey (void* Entry)
 /* Given a pointer to the user entry data, return a pointer to the index */
 {
-    return ((Macro*) Entry)->Name;
+    return &((Macro*) Entry)->Name;
 }
 
 
@@ -183,7 +184,7 @@ static int HT_Compare (const void* Key1, const void* Key2)
  * than zero if Key1 is greater then Key2.
  */
 {
-    return strcmp (Key1, Key2);
+    return SB_Compare (Key1, Key2);
 }
 
 
@@ -194,17 +195,16 @@ static int HT_Compare (const void* Key1, const void* Key2)
 
 
 
-static IdDesc* NewIdDesc (const char* Id)
+static IdDesc* NewIdDesc (const StrBuf* Id)
 /* Create a new IdDesc, initialize and return it */
 {
     /* Allocate memory */
-    unsigned Len = strlen (Id);
-    IdDesc* I = xmalloc (sizeof (IdDesc) + Len);
+    IdDesc* I = xmalloc (sizeof (IdDesc));
 
     /* Initialize the struct */
     I->Next = 0;
-    memcpy (I->Id, Id, Len);
-    I->Id [Len] = '\0';
+    I->Id   = AUTO_STRBUF_INITIALIZER;
+    SB_Copy (&I->Id, Id);
 
     /* Return the new struct */
     return I;
@@ -212,12 +212,11 @@ static IdDesc* NewIdDesc (const char* Id)
 
 
 
-static Macro* NewMacro (const char* Name, unsigned char Style)
+static Macro* NewMacro (const StrBuf* Name, unsigned char Style)
 /* Generate a new macro entry, initialize and return it */
 {
     /* Allocate memory */
-    unsigned Len = strlen (Name);
-    Macro* M = xmalloc (sizeof (Macro) + Len);
+    Macro* M = xmalloc (sizeof (Macro));
 
     /* Initialize the macro struct */
     InitHashNode (&M->Node, M);
@@ -229,7 +228,8 @@ static Macro* NewMacro (const char* Name, unsigned char Style)
     M->TokRoot    = 0;
     M->TokLast    = 0;
     M->Style	  = Style;
-    memcpy (M->Name, Name, Len+1);
+    M->Name       = AUTO_STRBUF_INITIALIZER;
+    SB_Copy (&M->Name, Name);
 
     /* Insert the macro into the global macro list */
     M->List = MacroRoot;
@@ -283,9 +283,15 @@ static void FreeMacExp (MacExp* E)
     /* One macro expansion less */
     --MacExpansions;
 
-    /* Free the parameter list */
+    /* Free the parameter lists */
     for (I = 0; I < E->ParamCount; ++I) {
-      	xfree (E->Params [I]);
+        /* Free one parameter list */
+        TokNode* N = E->Params[I];
+        while (N) {
+            TokNode* P = N->Next;
+            FreeTokNode (N);
+            N = P;
+        }
     }
     xfree (E->Params);
 
@@ -333,19 +339,26 @@ void MacDef (unsigned Style)
 	Error ("Identifier expected");
 	MacSkipDef (Style);
     	return;
+    } else if (!UbiquitousIdents && FindInstruction (&SVal) >= 0) {
+        /* The identifier is a name of a 6502 instruction, which is not
+         * allowed if not explicitly enabled.
+         */
+        Error ("Cannot use an instruction as macro name");
+        MacSkipDef (Style);
+        return;
     }
 
     /* Did we already define that macro? */
-    if (HT_Find (&MacroTab, SVal) != 0) {
+    if (HT_Find (&MacroTab, &SVal) != 0) {
        	/* Macro is already defined */
-     	Error ("A macro named `%s' is already defined", SVal);
+       	Error ("A macro named `%m%p' is already defined", &SVal);
      	/* Skip tokens until we reach the final .endmacro */
      	MacSkipDef (Style);
        	return;
     }
 
     /* Define the macro */
-    M = NewMacro (SVal, Style);
+    M = NewMacro (&SVal, Style);
 
     /* Switch to raw token mode and skip the macro name */
     EnterRawTokenMode ();
@@ -355,45 +368,45 @@ void MacDef (unsigned Style)
      * otherwise we may have parameters without braces.
      */
     if (Style == MAC_STYLE_CLASSIC) {
-	HaveParams = 1;
+     	HaveParams = 1;
     } else {
-	if (Tok == TOK_LPAREN) {
-	    HaveParams = 1;
-	    NextTok ();
-	} else {
-	    HaveParams = 0;
-	}
+     	if (Tok == TOK_LPAREN) {
+     	    HaveParams = 1;
+     	    NextTok ();
+     	} else {
+     	    HaveParams = 0;
+     	}
     }
 
     /* Parse the parameter list */
     if (HaveParams) {
 
-    	while (Tok == TOK_IDENT) {
+     	while (Tok == TOK_IDENT) {
 
-	    /* Create a struct holding the identifier */
-	    IdDesc* I = NewIdDesc (SVal);
+     	    /* Create a struct holding the identifier */
+     	    IdDesc* I = NewIdDesc (&SVal);
 
-	    /* Insert the struct into the list, checking for duplicate idents */
-	    if (M->ParamCount == 0) {
-		M->Params = I;
-	    } else {
-		IdDesc* List = M->Params;
-		while (1) {
-		    if (strcmp (List->Id, SVal) == 0) {
-			Error ("Duplicate symbol `%s'", SVal);
-		    }                                 
-		    if (List->Next == 0) {
-			break;
-		    } else {
-			List = List->Next;
-		    }
-		}
-		List->Next = I;
-	    }
-	    ++M->ParamCount;
+     	    /* Insert the struct into the list, checking for duplicate idents */
+     	    if (M->ParamCount == 0) {
+     		M->Params = I;
+     	    } else {
+     		IdDesc* List = M->Params;
+     		while (1) {
+     		    if (SB_Compare (&List->Id, &SVal) == 0) {
+     		   	Error ("Duplicate symbol `%m%p'", &SVal);
+     		    }
+     		    if (List->Next == 0) {
+     			break;
+     		    } else {
+     			List = List->Next;
+     		    }
+     		}
+     		List->Next = I;
+     	    }
+     	    ++M->ParamCount;
 
        	    /* Skip the name */
-	    NextTok ();
+     	    NextTok ();
 
 	    /* Maybe there are more params... */
 	    if (Tok == TOK_COMMA) {
@@ -450,14 +463,14 @@ void MacDef (unsigned Style)
        		NextTok ();
 
      		/* Need an identifer */
-     		if (Tok != TOK_IDENT) {
+     		if (Tok != TOK_IDENT && Tok != TOK_LOCAL_IDENT) {
      	       	    Error ("Identifier expected");
      		    SkipUntilSep ();
      		    break;
      		}
 
      		/* Put the identifier into the locals list and skip it */
-       	       	I = NewIdDesc (SVal);
+       	       	I = NewIdDesc (&SVal);
      		I->Next = M->Locals;
      		M->Locals = I;
      		++M->LocalCount;
@@ -483,7 +496,7 @@ void MacDef (unsigned Style)
      	    unsigned Count = 0;
      	    IdDesc* I = M->Params;
      	    while (I) {
-     	       	if (strcmp (I->Id, SVal) == 0) {
+     	       	if (SB_Compare (&I->Id, &SVal) == 0) {
      	       	    /* Local param name, replace it */
      	       	    T->Tok  = TOK_MACPARAM;
      	       	    T->IVal = Count;
@@ -588,20 +601,26 @@ static int MacExpand (void* Data)
        	}
 
        	/* If it's an identifier, it may in fact be a local symbol */
-       	if (Tok == TOK_IDENT && Mac->M->LocalCount) {
+       	if ((Tok == TOK_IDENT || Tok == TOK_LOCAL_IDENT) && Mac->M->LocalCount) {
        	    /* Search for the local symbol in the list */
        	    unsigned Index = 0;
        	    IdDesc* I = Mac->M->Locals;
        	    while (I) {
-       	       	if (strcmp (SVal, I->Id) == 0) {
+       	       	if (SB_Compare (&SVal, &I->Id) == 0) {
        	       	    /* This is in fact a local symbol, change the name. Be sure
                      * to generate a local label name if the original name was
                      * a local label, and also generate a name that cannot be
                      * generated by a user.
                      */
-                    unsigned PrefixLen = (I->Id[0] == LocalStart);
-       	       	    sprintf (SVal, "%.*sLOCAL-MACRO-SYMBOL-%04X", PrefixLen,
-                             I->Id, Mac->LocalStart + Index);
+                    if (SB_At (&I->Id, 0) == LocalStart) {
+                        /* Must generate a local symbol */
+                        SB_Printf (&SVal, "%cLOCAL-MACRO_SYMBOL-%04X",
+                                   LocalStart, Mac->LocalStart + Index);
+                    } else {
+                        /* Global symbol */
+                        SB_Printf (&SVal, "LOCAL-MACRO_SYMBOL-%04X",
+                                   Mac->LocalStart + Index);
+                    }
        	       	    break;
        	       	}
        	       	/* Next symbol */
@@ -647,7 +666,9 @@ MacEnd:
 static void StartExpClassic (Macro* M)
 /* Start expanding the classic macro M */
 {
-    MacExp* E;
+    MacExp*     E;
+    Token       Term;
+
 
     /* Skip the macro name */
     NextTok ();
@@ -662,21 +683,24 @@ static void StartExpClassic (Macro* M)
 
        	/* Check for maximum parameter count */
 	if (E->ParamCount >= M->ParamCount) {
-	    Error ("Too many macro parameters");
-	    SkipUntilSep ();                    
+       	    ErrorSkip ("Too many macro parameters");
 	    break;
 	}
 
-	/* Read tokens for one parameter, accept empty params */
+        /* The macro may optionally be enclosed in curly braces */
+        Term = GetTokListTerm (TOK_COMMA);
+
+       	/* Read tokens for one parameter, accept empty params */
 	Last = 0;
-	while (Tok != TOK_COMMA && Tok != TOK_SEP) {
+	while (Tok != Term && Tok != TOK_SEP) {
 
 	    TokNode* T;
 
 	    /* Check for end of file */
 	    if (Tok == TOK_EOF) {
-	    	Error ("Unexpected end of file");
-	    	return;
+	     	Error ("Unexpected end of file");
+                FreeMacExp (E);
+	     	return;
 	    }
 
 	    /* Get the next token in a node */
@@ -686,7 +710,7 @@ static void StartExpClassic (Macro* M)
 	    if (Last == 0) {
 	      	E->Params [E->ParamCount] = T;
 	    } else {
-	    	Last->Next = T;
+	     	Last->Next = T;
 	    }
 	    Last = T;
 
@@ -697,14 +721,28 @@ static void StartExpClassic (Macro* M)
 	/* One parameter more */
 	++E->ParamCount;
 
+        /* If the macro argument was enclosed in curly braces, end-of-line
+         * is an error. Skip the closing curly brace.
+         */
+        if (Term == TOK_RCURLY) {
+            if (Tok == TOK_SEP) {
+                Error ("End of line encountered within macro argument");
+                break;
+            }
+            NextTok ();
+        }
+
 	/* Check for a comma */
 	if (Tok == TOK_COMMA) {
-	    NextTok ();
-	} else {
-	    break;
-	}
+ 	    NextTok ();
+ 	} else {
+ 	    break;
+ 	}
     }
 
+    /* We must be at end of line now, otherwise something is wrong */
+    ExpectSep ();
+
     /* Insert a new token input function */
     PushInput (MacExpand, E, ".MACRO");
 }
@@ -728,14 +766,17 @@ static void StartExpDefine (Macro* M)
     /* Read the actual parameters */
     while (Count--) {
 
-       	TokNode* Last;
+       	TokNode*   Last;
+
+        /* The macro may optionally be enclosed in curly braces */
+        Token Term = GetTokListTerm (TOK_COMMA);
 
        	/* Check if there is really a parameter */
-       	if (TokIsSep (Tok) || Tok == TOK_COMMA) {
-       	    Error ("Macro parameter expected");
-       	    SkipUntilSep ();
-       	    return;
-       	}
+       	if (TokIsSep (Tok) || Tok == Term) {
+            ErrorSkip ("Macro parameter #%u is empty", E->ParamCount+1);
+            FreeMacExp (E);
+            return;
+        }
 
        	/* Read tokens for one parameter */
        	Last = 0;
@@ -757,11 +798,22 @@ static void StartExpDefine (Macro* M)
 	    /* And skip it... */
 	    NextTok ();
 
-	} while (Tok != TOK_COMMA && !TokIsSep (Tok));
+       	} while (Tok != Term && !TokIsSep (Tok));
 
 	/* One parameter more */
 	++E->ParamCount;
 
+        /* If the macro argument was enclosed in curly braces, end-of-line
+         * is an error. Skip the closing curly brace.
+         */
+        if (Term == TOK_RCURLY) {
+            if (TokIsSep (Tok)) {
+                Error ("End of line encountered within macro argument");
+                break;
+            }
+            NextTok ();
+        }
+
        	/* Check for a comma */
        	if (Count > 0) {
        	    if (Tok == TOK_COMMA) {
@@ -774,7 +826,7 @@ static void StartExpDefine (Macro* M)
 
     /* Macro expansion will overwrite the current token. This is a problem
      * for define style macros since these are called from the scanner level.
-     * To avoid it, remember the current token and re-insert it if macro
+     * To avoid it, remember the current token and re-insert it, once macro
      * expansion is done.
      */
     E->Final = NewTokNode ();
@@ -789,7 +841,7 @@ void MacExpandStart (void)
 /* Start expanding the macro in SVal */
 {
     /* Search for the macro */
-    Macro* M = HT_FindEntry (&MacroTab, SVal);
+    Macro* M = HT_FindEntry (&MacroTab, &SVal);
     CHECK (M != 0);
 
     /* Call the apropriate subroutine */
@@ -814,7 +866,7 @@ void MacAbort (void)
 
 
 
-int IsMacro (const char* Name)
+int IsMacro (const StrBuf* Name)
 /* Return true if the given name is the name of a macro */
 {
     return (HT_Find (&MacroTab, Name) != 0);
@@ -822,7 +874,7 @@ int IsMacro (const char* Name)
 
 
 
-int IsDefine (const char* Name)
+int IsDefine (const StrBuf* Name)
 /* Return true if the given name is the name of a define style macro */
 {
     Macro* M = HT_FindEntry (&MacroTab, Name);
@@ -839,6 +891,3 @@ int InMacExpansion (void)
 
 
 
-
-
-
