@@ -6,10 +6,10 @@
 /*                                                                           */
 /*                                                                           */
 /*                                                                           */
-/* (C) 2001-2002 Ullrich von Bassewitz                                       */
-/*               Wacholderweg 14                                             */
-/*               D-70597 Stuttgart                                           */
-/* EMail:        uz@cc65.org                                                 */
+/* (C) 2001-2009, Ullrich von Bassewitz                                      */
+/*                Roemerstrasse 52                                           */
+/*                D-70794 Filderstadt                                        */
+/* EMail:         uz@cc65.org                                                */
 /*                                                                           */
 /*                                                                           */
 /* This software is provided 'as-is', without any expressed or implied       */
@@ -39,6 +39,7 @@
 /* common */
 #include "chartype.h"
 #include "coll.h"
+#include "debugflag.h"
 
 /* cc65 */
 #include "codeent.h"
@@ -67,11 +68,15 @@ static const char CmpSuffixTab [][4] = {
  */
 typedef struct FuncInfo FuncInfo;
 struct FuncInfo {
-    const char*	    Name;	/* Function name */
-    unsigned short  Use;	/* Register usage */
-    unsigned short  Chg;   	/* Changed/destroyed registers */
+    const char*	    Name;    	/* Function name */
+    unsigned short  Use;     	/* Register usage */
+    unsigned short  Chg;     	/* Changed/destroyed registers */
 };
 
+/* Note for the shift functions: Shifts are done modulo 32, so all shift
+ * routines are marked to use only the A register. The remainder is ignored
+ * anyway.
+ */
 static const FuncInfo FuncInfoTable[] = {
     { "addeq0sp",       REG_AX,               REG_AXY                        },
     { "addeqysp",       REG_AXY,              REG_AXY                        },
@@ -80,6 +85,18 @@ static const FuncInfo FuncInfoTable[] = {
     { "aslax2",         REG_AX,	       	      REG_AX | REG_TMP1		     },
     { "aslax3",         REG_AX,	       	      REG_AX | REG_TMP1		     },
     { "aslax4",         REG_AX,	       	      REG_AX | REG_TMP1		     },
+    { "asleax1",        REG_EAX,              REG_EAX | REG_TMP1       	     },
+    { "asleax2",        REG_EAX,              REG_EAX | REG_TMP1	     },
+    { "asleax3",        REG_EAX,              REG_EAX | REG_TMP1	     },
+    { "asleax4",        REG_EAX,              REG_EAXY | REG_TMP1    	     },
+    { "asrax1",         REG_AX,	       	      REG_AX | REG_TMP1	 	     },
+    { "asrax2",         REG_AX,	       	      REG_AX | REG_TMP1	 	     },
+    { "asrax3",         REG_AX,	       	      REG_AX | REG_TMP1	 	     },
+    { "asrax4",         REG_AX,	       	      REG_AX | REG_TMP1	 	     },
+    { "asreax1",        REG_EAX,              REG_EAX | REG_TMP1       	     },
+    { "asreax2",        REG_EAX,              REG_EAX | REG_TMP1 	     },
+    { "asreax3",        REG_EAX,              REG_EAX | REG_TMP1 	     },
+    { "asreax4",        REG_EAX,              REG_EAXY | REG_TMP1      	     },
     { "bnega",          REG_A,                REG_AX			     },
     { "bnegax",         REG_AX,               REG_AX			     },
     { "bnegeax",        REG_EAX,              REG_EAX			     },
@@ -87,12 +104,13 @@ static const FuncInfo FuncInfoTable[] = {
     { "boolge",	       	REG_NONE,      	      REG_AX			     },
     { "boolgt",	       	REG_NONE,      	      REG_AX			     },
     { "boolle",	       	REG_NONE,      	      REG_AX			     },
-    { "boollt",	       	REG_NONE,      	      REG_AX			     },
+    { "boollt",	       	REG_NONE,      	      REG_AX		      	     },
     { "boolne",	       	REG_NONE,      	      REG_AX			     },
     { "booluge",       	REG_NONE,      	      REG_AX			     },
     { "boolugt",       	REG_NONE,      	      REG_AX			     },
     { "boolule",       	REG_NONE,      	      REG_AX			     },
     { "boolult",       	REG_NONE,      	      REG_AX			     },
+    { "callax",         REG_AX,               REG_ALL                        },
     { "complax",        REG_AX,               REG_AX			     },
     { "decax1",        	REG_AX,	       	      REG_AX			     },
     { "decax2",        	REG_AX,	       	      REG_AX			     },
@@ -112,8 +130,15 @@ static const FuncInfo FuncInfoTable[] = {
     { "decsp6",        	REG_NONE,      	      REG_A			     },
     { "decsp7",        	REG_NONE,      	      REG_A			     },
     { "decsp8",        	REG_NONE,      	      REG_A			     },
-    { "incax1",         REG_AX,               REG_AX			     },
-    { "incax2",         REG_AX,               REG_AX			     },
+    { "incax1",         REG_AX,               REG_AX       		     },
+    { "incax2",         REG_AX,               REG_AX  			     },
+    { "incax3",         REG_AX,               REG_AXY | REG_TMP1             },
+    { "incax4",         REG_AX,               REG_AXY | REG_TMP1  	     },
+    { "incax5",         REG_AX,               REG_AXY | REG_TMP1 	     },
+    { "incax6",         REG_AX,               REG_AXY | REG_TMP1 	     },
+    { "incax7",         REG_AX,               REG_AXY | REG_TMP1 	     },
+    { "incax8",         REG_AX,               REG_AXY | REG_TMP1 	     },
+    { "incaxy",         REG_AXY,              REG_AXY | REG_TMP1       	     },
     { "incsp1",	       	REG_NONE,      	      REG_NONE			     },
     { "incsp2",	       	REG_NONE,      	      REG_Y			     },
     { "incsp3",	       	REG_NONE,      	      REG_Y			     },
@@ -137,14 +162,22 @@ static const FuncInfo FuncInfoTable[] = {
     { "ldeaxi",         REG_AX,               REG_EAXY | REG_PTR1            },
     { "ldeaxidx",       REG_AXY,              REG_EAXY | REG_PTR1            },
     { "ldeaxysp",       REG_Y,                REG_EAXY                       },
-    { "leaasp",         REG_A,                REG_AX			     },
+    { "leaa0sp",        REG_A,                REG_AX			     },
+    { "leaaxsp",        REG_AX,               REG_AX			     },
     { "lsubeq",         REG_EAXY|REG_PTR1_LO, REG_EAXY | REG_PTR1_HI         },
     { "lsubeq0sp",      REG_EAX,              REG_EAXY                       },
     { "lsubeq1",        REG_Y | REG_PTR1_LO,  REG_EAXY | REG_PTR1_HI         },
     { "lsubeqa",        REG_AY | REG_PTR1_LO, REG_EAXY | REG_PTR1_HI         },
     { "lsubeqysp",      REG_EAXY,             REG_EAXY                       },
+    { "mulax10",        REG_AX,               REG_AX | REG_PTR1              },
+    { "mulax3",         REG_AX,               REG_AX | REG_PTR1              },
+    { "mulax5",         REG_AX,               REG_AX | REG_PTR1              },
+    { "mulax6",         REG_AX,               REG_AX | REG_PTR1              },
+    { "mulax7",         REG_AX,               REG_AX | REG_PTR1              },
+    { "mulax9",         REG_AX,               REG_AX | REG_PTR1              },
     { "negax",          REG_AX,               REG_AX			     },
     { "push0", 	       	REG_NONE,             REG_AXY			     },
+    { "push0ax",        REG_AX,               REG_Y | REG_SREG               },
     { "push1", 	       	REG_NONE,             REG_AXY			     },
     { "push2", 	       	REG_NONE,             REG_AXY			     },
     { "push3", 	       	REG_NONE,             REG_AXY			     },
@@ -155,7 +188,8 @@ static const FuncInfo FuncInfoTable[] = {
     { "pusha", 	       	REG_A, 	       	      REG_Y			     },
     { "pusha0",	       	REG_A, 	       	      REG_XY			     },
     { "pusha0sp",       REG_NONE,             REG_AY                         },
-    { "pushax",	       	REG_AX,	       	      REG_Y			     },
+    { "pushaFF",       	REG_A,	       	      REG_Y  			     },
+    { "pushax",	       	REG_AX,	       	      REG_Y  			     },
     { "pushaysp",       REG_Y,                REG_AY                         },
     { "pushc0",	       	REG_NONE,             REG_A | REG_Y                  },
     { "pushc1",	       	REG_NONE,             REG_A | REG_Y                  },
@@ -168,54 +202,129 @@ static const FuncInfo FuncInfoTable[] = {
     { "regswap",        REG_AXY,              REG_AXY | REG_TMP1             },
     { "regswap1",       REG_XY,               REG_A                          },
     { "regswap2",       REG_XY,               REG_A | REG_Y                  },
+    { "return0",        REG_NONE,             REG_AX                         },
+    { "return1",        REG_NONE,             REG_AX                         },
     { "shlax1",         REG_AX,	       	      REG_AX | REG_TMP1		     },
     { "shlax2",         REG_AX,	       	      REG_AX | REG_TMP1		     },
     { "shlax3",         REG_AX,	       	      REG_AX | REG_TMP1		     },
     { "shlax4",         REG_AX,	       	      REG_AX | REG_TMP1		     },
+    { "shleax1",        REG_EAX,       	      REG_EAX | REG_TMP1       	     },
+    { "shleax2",        REG_EAX,       	      REG_EAX | REG_TMP1	     },
+    { "shleax3",        REG_EAX,       	      REG_EAX | REG_TMP1	     },
+    { "shleax4",        REG_EAX,       	      REG_EAXY | REG_TMP1      	     },
     { "shrax1",         REG_AX,               REG_AX | REG_TMP1		     },
     { "shrax2",         REG_AX,               REG_AX | REG_TMP1		     },
     { "shrax3",         REG_AX,               REG_AX | REG_TMP1		     },
     { "shrax4",         REG_AX,               REG_AX | REG_TMP1		     },
-    { "shreax1",        REG_EAX,              REG_AX | REG_TMP1		     },
-    { "shreax2",        REG_EAX,              REG_AX | REG_TMP1		     },
-    { "shreax3",        REG_EAX,              REG_AX | REG_TMP1		     },
-    { "shreax4",        REG_EAX,              REG_AX | REG_TMP1		     },
+    { "shreax1",        REG_EAX,              REG_EAX | REG_TMP1       	     },
+    { "shreax2",        REG_EAX,              REG_EAX | REG_TMP1	     },
+    { "shreax3",        REG_EAX,              REG_EAX | REG_TMP1	     },
+    { "shreax4",        REG_EAX,              REG_EAXY | REG_TMP1      	     },
     { "staspidx",       REG_A | REG_Y,        REG_Y | REG_TMP1 | REG_PTR1    },
     { "stax0sp",        REG_AX,               REG_Y   			     },
+    { "staxspidx",      REG_AXY,              REG_TMP1 | REG_PTR1            },
     { "staxysp",        REG_AXY,              REG_Y   			     },
     { "steax0sp",       REG_EAX,              REG_Y                          },
     { "steaxysp",       REG_EAXY,             REG_Y                          },
     { "subeq0sp",       REG_AX,               REG_AXY                        },
     { "subeqysp",       REG_AXY,              REG_AXY                        },
+    { "tosadd0ax",      REG_AX,               REG_EAXY | REG_TMP1            },
     { "tosadda0",       REG_A,                REG_AXY                        },
     { "tosaddax",       REG_AX,               REG_AXY                        },
+    { "tosaddeax",      REG_EAX,              REG_EAXY | REG_TMP1            },
+    { "tosand0ax",      REG_AX,               REG_EAX | REG_TMP1             },
     { "tosanda0",       REG_A,                REG_AXY                        },
     { "tosandax",       REG_AX,               REG_AXY                        },
-    { "tosdiva0",       REG_AY,       	      REG_ALL 			     },
-    { "tosdivax",       REG_AXY,              REG_ALL 			     },
-    { "tosdiveax",      REG_EAXY,             REG_ALL 			     },
+    { "tosandeax",      REG_EAX,              REG_EAX | REG_TMP1             },
+    { "tosaslax",       REG_A,                REG_AXY | REG_TMP1             },
+    { "tosasleax",      REG_A,                REG_EAXY | REG_TMP1            },
+    { "tosasrax",       REG_A,                REG_AXY | REG_TMP1             },
+    { "tosasreax",      REG_A,                REG_EAXY | REG_TMP1            },
+    { "tosdiv0ax",      REG_AX,               REG_ALL 			     },
+    { "tosdiva0",       REG_A,                REG_ALL 			     },
+    { "tosdivax",       REG_AX,               REG_ALL 			     },
+    { "tosdiveax",      REG_EAX,              REG_ALL 			     },
+    { "toseq00",        REG_NONE,      	      REG_AXY | REG_SREG             },
+    { "toseqa0",        REG_A,        	      REG_AXY | REG_SREG             },
+    { "toseqax",        REG_AX,        	      REG_AXY | REG_SREG             },
     { "toseqeax",       REG_EAX,       	      REG_AXY | REG_PTR1             },
+    { "tosge00",        REG_NONE,      	      REG_AXY | REG_SREG             },
+    { "tosgea0",        REG_A,        	      REG_AXY | REG_SREG             },
+    { "tosgeax",        REG_AX,        	      REG_AXY | REG_SREG             },
     { "tosgeeax",       REG_EAX,       	      REG_AXY | REG_PTR1             },
+    { "tosgt00",        REG_NONE,      	      REG_AXY | REG_SREG             },
+    { "tosgta0",        REG_A,        	      REG_AXY | REG_SREG             },
+    { "tosgtax",        REG_AX,        	      REG_AXY | REG_SREG             },
     { "tosgteax",       REG_EAX,       	      REG_AXY | REG_PTR1             },
     { "tosicmp",       	REG_AX,	       	      REG_AXY | REG_SREG	     },
+    { "tosicmp0",       REG_A,                REG_AXY | REG_SREG             },
     { "toslcmp",        REG_EAX,       	      REG_A | REG_Y | REG_PTR1       },
+    { "tosle00",        REG_NONE,      	      REG_AXY | REG_SREG             },
+    { "toslea0",        REG_A,        	      REG_AXY | REG_SREG             },
+    { "tosleax",        REG_AX,        	      REG_AXY | REG_SREG             },
     { "tosleeax",       REG_EAX,       	      REG_AXY | REG_PTR1             },
+    { "toslt00",        REG_NONE,      	      REG_AXY | REG_SREG             },
+    { "toslta0",        REG_A,        	      REG_AXY | REG_SREG             },
+    { "tosltax",        REG_AX,        	      REG_AXY | REG_SREG             },
     { "toslteax",       REG_EAX,       	      REG_AXY | REG_PTR1             },
-    { "tosmula0",       REG_AX,	       	      REG_ALL 			     },
+    { "tosmod0ax",      REG_AX,               REG_ALL 			     },
+    { "tosmodeax",      REG_EAX,              REG_ALL 			     },
+    { "tosmul0ax",      REG_AX,       	      REG_ALL 			     },
+    { "tosmula0",       REG_A,	       	      REG_ALL 			     },
     { "tosmulax",       REG_AX,	       	      REG_ALL 			     },
     { "tosmuleax",      REG_EAX,       	      REG_ALL 			     },
+    { "tosne00",        REG_NONE,      	      REG_AXY | REG_SREG             },
+    { "tosnea0",        REG_A,        	      REG_AXY | REG_SREG             },
+    { "tosneax",        REG_AX,        	      REG_AXY | REG_SREG             },
     { "tosneeax",       REG_EAX,       	      REG_AXY | REG_PTR1             },
-    { "tosshreax",      REG_EAX,       	      REG_EAXY | REG_PTR1 | REG_PTR2 },
+    { "tosor0ax",       REG_AX,               REG_EAX | REG_TMP1             },
+    { "tosora0",        REG_A,        	      REG_AXY | REG_TMP1             },
+    { "tosorax",        REG_AX,        	      REG_AXY | REG_TMP1             },
+    { "tosoreax",       REG_EAX,              REG_EAX | REG_TMP1             },
+    { "tosrsub0ax",     REG_AX,               REG_EAX | REG_TMP1             },
+    { "tosrsuba0",      REG_A,                REG_AX | REG_TMP1              },
+    { "tosrsubax",      REG_AX,               REG_AX | REG_TMP1              },
+    { "tosrsubeax",     REG_EAX,              REG_EAX | REG_TMP1             },
+    { "tosshlax",       REG_A,                REG_AXY | REG_TMP1             },
+    { "tosshleax",      REG_A,         	      REG_EAXY | REG_TMP1            },
+    { "tosshrax",       REG_A,                REG_AXY | REG_TMP1             },
+    { "tosshreax",      REG_A,         	      REG_EAXY | REG_TMP1            },
+    { "tossub0ax",      REG_AX,               REG_EAXY                       },
     { "tossuba0",       REG_A,                REG_AXY                        },
     { "tossubax",       REG_AX,               REG_AXY                        },
     { "tossubeax",      REG_EAX,              REG_EAXY                       },
+    { "tosudiv0ax",     REG_AX,               REG_ALL & ~REG_SAVE            },
+    { "tosudiva0",      REG_A,                REG_EAXY | REG_PTR1            }, /* also ptr4 */
+    { "tosudivax",      REG_AX,               REG_EAXY | REG_PTR1            }, /* also ptr4 */
+    { "tosudiveax",     REG_EAX,              REG_ALL & ~REG_SAVE            },
+    { "tosuge00",       REG_NONE,      	      REG_AXY | REG_SREG             },
+    { "tosugea0",       REG_A,        	      REG_AXY | REG_SREG             },
+    { "tosugeax",       REG_AX,        	      REG_AXY | REG_SREG             },
     { "tosugeeax",      REG_EAX,       	      REG_AXY | REG_PTR1             },
+    { "tosugt00",       REG_NONE,      	      REG_AXY | REG_SREG             },
+    { "tosugta0",       REG_A,        	      REG_AXY | REG_SREG             },
+    { "tosugtax",       REG_AX,        	      REG_AXY | REG_SREG             },
     { "tosugteax",      REG_EAX,       	      REG_AXY | REG_PTR1             },
+    { "tosule00",       REG_NONE,      	      REG_AXY | REG_SREG             },
+    { "tosulea0",       REG_A,        	      REG_AXY | REG_SREG             },
+    { "tosuleax",       REG_AX,        	      REG_AXY | REG_SREG             },
     { "tosuleeax",      REG_EAX,       	      REG_AXY | REG_PTR1             },
+    { "tosult00",       REG_NONE,      	      REG_AXY | REG_SREG             },
+    { "tosulta0",       REG_A,        	      REG_AXY | REG_SREG             },
+    { "tosultax",       REG_AX,        	      REG_AXY | REG_SREG             },
     { "tosulteax",      REG_EAX,       	      REG_AXY | REG_PTR1             },
-    { "tosumula0",      REG_AX,	       	      REG_ALL			     },
+    { "tosumod0ax",     REG_AX,               REG_ALL & ~REG_SAVE            },
+    { "tosumoda0",      REG_A,                REG_EAXY | REG_PTR1            }, /* also ptr4 */
+    { "tosumodax",      REG_AX,               REG_EAXY | REG_PTR1            }, /* also ptr4 */
+    { "tosumodeax",     REG_EAX,              REG_ALL & ~REG_SAVE            },
+    { "tosumul0ax",     REG_AX,       	      REG_ALL			     },
+    { "tosumula0",      REG_A,                REG_ALL			     },
     { "tosumulax",      REG_AX,	       	      REG_ALL			     },
     { "tosumuleax",     REG_EAX,       	      REG_ALL			     },
+    { "tosxor0ax",      REG_AX,               REG_EAX | REG_TMP1             },
+    { "tosxora0",       REG_A,                REG_AX | REG_TMP1              },
+    { "tosxorax",       REG_AX,               REG_AX | REG_TMP1              },
+    { "tosxoreax",      REG_EAX,              REG_EAX | REG_TMP1             },
     { "tsteax",         REG_EAX,              REG_Y                          },
     { "utsteax",        REG_EAX,              REG_Y                          },
 };
@@ -278,14 +387,15 @@ void GetFuncInfo (const char* Name, unsigned short* Use, unsigned short* Chg)
      	/* Did we find it in the top level table? */
      	if (E && IsTypeFunc (E->Type)) {
 
+            FuncDesc* D = E->V.F.Func;
+
      	    /* A function may use the A or A/X registers if it is a fastcall
      	     * function. If it is not a fastcall function but a variadic one,
 	     * it will use the Y register (the parameter size is passed here).
 	     * In all other cases, no registers are used. However, we assume
 	     * that any function will destroy all registers.
      	     */
-     	    FuncDesc* D = E->V.F.Func;
-     	    if ((D->Flags & FD_FASTCALL) != 0 && D->ParamCount > 0) {
+     	    if (IsQualFastcall (E->Type) && D->ParamCount > 0) {
      		/* Will use registers depending on the last param */
                 unsigned LastParamSize = CheckedSizeOf (D->LastParam->Type);
      		if (LastParamSize == 1) {
@@ -330,8 +440,19 @@ void GetFuncInfo (const char* Name, unsigned short* Use, unsigned short* Chg)
 	    /* Use the information we have */
 	    *Use = Info->Use;
 	    *Chg = Info->Chg;
-	    return;
-	}
+	} else {
+            /* It's an internal function we have no information for. If in
+             * debug mode, output an additional warning, so we have a chance
+             * to fix it. Otherwise assume that the internal function will
+             * use and change all registers.
+             */
+            if (Debug) {
+                fprintf (stderr, "No info about internal function `%s'\n", Name);
+            }
+            *Use = REG_ALL;
+            *Chg = REG_ALL;
+        }
+        return;
     }
 
     /* Function not found - assume that the primary register is input, and all
@@ -409,7 +530,7 @@ static unsigned GetRegInfo2 (CodeSeg* S,
 	/* Evaluate the used registers */
 	R = E->Use;
 	if (E->OPC == OP65_RTS ||
-	    ((E->Info & OF_BRA) != 0 && E->JumpTo == 0)) {
+       	    ((E->Info & OF_UBRA) != 0 && E->JumpTo == 0)) {
 	    /* This instruction will leave the function */
 	    R |= S->ExitRegs;
 	}
@@ -465,30 +586,37 @@ static unsigned GetRegInfo2 (CodeSeg* S,
 	 */
 	} else if ((E->Info & OF_CBRA) != 0) {
 
+            /* Recursively determine register usage at the branch target */
+	    unsigned U1;
+	    unsigned U2;
+
     	    if (E->JumpTo) {
 
-	       	/* Recursively determine register usage at the branch target */
-		unsigned U1;
-		unsigned U2;
-
+                /* Jump to internal label */
 		U1 = GetRegInfo2 (S, E->JumpTo->Owner, -1, Visited, Used, Unused, Wanted);
-		if (U1 == REG_ALL) {
-		    /* All registers used, no need for second call */
-		    return REG_AXY;
-		}
-		if (Index < 0) {
-		    Index = CS_GetEntryIndex (S, E);
-		}
-       	       	if ((E = CS_GetEntry (S, ++Index)) == 0) {
-		    Internal ("GetRegInfo2: No next entry!");
-		}
-       	       	U2 = GetRegInfo2 (S, E, Index, Visited, Used, Unused, Wanted);
-	   	return U1 | U2;	       	/* Used in any of the branches */
 
-	    } else {
-	   	/* Jump to global symbol */
-	  	break;
-	    }
+            } else {
+
+                /* Jump to external label. This will effectively exit the
+                 * function, so we use the exitregs information here.
+                 */
+                U1 = S->ExitRegs;
+
+            }
+
+            /* Get the next entry */
+            if (Index < 0) {
+                Index = CS_GetEntryIndex (S, E);
+            }
+            if ((E = CS_GetEntry (S, ++Index)) == 0) {
+                Internal ("GetRegInfo2: No next entry!");
+            }
+
+            /* Follow flow if branch not taken */
+            U2 = GetRegInfo2 (S, E, Index, Visited, Used, Unused, Wanted);
+
+            /* Registers are used if they're use in any of the branches */
+            return U1 | U2;
 
        	} else {
 
